@@ -85,8 +85,10 @@ def get_acse_residual_fqe(fqe_wf: Wavefunction, fqe_ham: RestrictedHamiltonian,
 
     Args:
         fqe_wf:  fqe.Wavefunction object to calculate expectation value with
+
         fqe_ham: fqe.RestrictedHamiltonian operator corresponding to a chemical
                  Hamiltonian
+
         norbs: Number of orbitals. Number of spatial orbitals
 
     Returns:
@@ -181,8 +183,10 @@ def get_tpdm_grad_fqe(fqe_wf, acse_res_tensor, norbs):
 
     Args:
         fqe_wf:  fqe.Wavefunction object to calculate expectation value with
+
         acse_res_tensor: fqe.RestrictedHamiltonian operator corresponding to a chemical
                          Hamiltonian
+
         norbs: Number of orbitals. Number of spatial orbitals
 
     Returns:
@@ -371,8 +375,10 @@ def get_acse_residual_fqe_parallel(fqe_wf, fqe_ham, norbs):
 
     Args:
         fqe_wf:  fqe.Wavefunction object to calculate expectation value with
+
         fqe_ham: fqe.RestrictedHamiltonian operator corresponding to a chemical
                  Hamiltonian
+
         norbs: Number of orbitals. Number of spatial orbitals
 
     Returns:
@@ -490,8 +496,10 @@ def get_tpdm_grad_fqe_parallel(fqe_wf, acse_res_tensor, norbs):
 
     Args:
         fqe_wf:  fqe.Wavefunction object to calculate expectation value with
+
         fqe_ham: fqe.RestrictedHamiltonian operator corresponding to a chemical
                  Hamiltonian
+
         norbs: Number of orbitals. Number of spatial orbitals
 
     Returns:
@@ -574,7 +582,9 @@ def two_rdo_commutator(two_body_tensor: np.ndarray, tpdm: np.ndarray,
 
     Args:
         two_body_tensor: 4-tensor for the coefficients of A
+
         tpdm: spin-orbital two-RDM p^ q^ r s corresponding to (1'2'2 1)
+
         d3: spin-orbital three-RDM p^ q^ r^ s t u corresponding to (1'2'3'32 1)
     """
     dim = tpdm.shape[0]
@@ -668,39 +678,32 @@ def two_rdo_commutator_symm(two_body_tensor: np.ndarray, tpdm: np.ndarray,
 
     Args:
         two_body_tensor: 4-tensor for the coefficients of A
+
         tpdm: spin-orbital two-RDM p^ q^ r s corresponding to (1'2'2 1)
+
         d3: spin-orbital three-RDM p^ q^ r^ s t u corresponding to (1'2'3'32 1)
     """
     dim = tpdm.shape[0]
-    tensor_of_expectation = np.zeros(tuple([dim] * 4), dtype=tpdm.dtype)
     k2 = two_body_tensor.transpose(0, 1, 3, 2)
-    for p, q, r, s in product(range(dim), repeat=4):
-        commutator_expectation = 0.
-        #   (  -2.00000) k2(p,q,a,b) cre(a) cre(b) des(r) des(s)
-        commutator_expectation += -2. * np.einsum('ab,ab', k2[p, q, :, :],
-                                                  tpdm[:, :, r, s])
+    tensor_of_expectation = np.zeros(tuple([dim] * 4), dtype=tpdm.dtype)
+    #   (  -2.00000) k2(p,q,a,b) cre(a) cre(b) des(r) des(s)
+    tensor_of_expectation += -2. * np.einsum('pqab,abrs->pqrs', k2, tpdm)
 
-        #   (   2.00000) k2(r,s,a,b) cre(p) cre(q) des(a) des(b)
-        commutator_expectation += 2. * np.einsum('ab,ab', k2[r, s, :, :],
-                                                 tpdm[p, q, :, :])
+    #   (   2.00000) k2(r,s,a,b) cre(p) cre(q) des(a) des(b)
+    tensor_of_expectation += 2. * np.einsum('rsab,pqab->pqrs', k2, tpdm)
 
-        #   (   2.00000) k2(p,a,b,c) cre(q) cre(b) cre(c) des(r) des(s) des(a)
-        commutator_expectation += 2. * np.einsum('abc,bca', k2[p, :, :, :],
-                                                 d3[q, :, :, r, s, :])
+    #   (   2.00000) k2(p,a,b,c) cre(q) cre(b) cre(c) des(r) des(s) des(a)
+    tensor_of_expectation += 2. * np.einsum('pabc,qbcrsa->pqrs', k2, d3)
 
-        #   (  -2.00000) k2(q,a,b,c) cre(p) cre(b) cre(c) des(r) des(s) des(a)
-        commutator_expectation += -2. * np.einsum('abc,bca', k2[q, :, :, :],
-                                                  d3[p, :, :, r, s, :])
+    #   (  -2.00000) k2(q,a,b,c) cre(p) cre(b) cre(c) des(r) des(s) des(a)
+    tensor_of_expectation += -2. * np.einsum('qabc,pbcrsa->pqrs', k2, d3)
 
-        #   (  -2.00000) k2(r,a,b,c) cre(p) cre(q) cre(a) des(s) des(b) des(c)
-        commutator_expectation += -2. * np.einsum('abc,abc', k2[r, :, :, :],
-                                                  d3[p, q, :, s, :, :])
+    #   (  -2.00000) k2(r,a,b,c) cre(p) cre(q) cre(a) des(s) des(b) des(c)
+    tensor_of_expectation += -2. * np.einsum('rabc,pqasbc->pqrs', k2, d3)
 
-        #   (   2.00000) k2(s,a,b,c) cre(p) cre(q) cre(a) des(r) des(b) des(c)
-        commutator_expectation += 2. * np.einsum('abc,abc', k2[s, :, :, :],
-                                                 d3[p, q, :, r, :, :])
+    #   (   2.00000) k2(s,a,b,c) cre(p) cre(q) cre(a) des(r) des(b) des(c)
+    tensor_of_expectation += 2. * np.einsum('sabc,pqarbc->pqrs', k2, d3)
 
-        tensor_of_expectation[p, q, r, s] = commutator_expectation
     return tensor_of_expectation
 
 
@@ -715,7 +718,9 @@ def two_rdo_commutator_antisymm(two_body_tensor: np.ndarray, tpdm: np.ndarray,
 
     Args:
         two_body_tensor: 4-tensor for the coefficients of A
+
         tpdm: spin-orbital two-RDM p^ q^ r s corresponding to (1'2'2 1)
+
         d3: spin-orbital three-RDM p^ q^ r^ s t u corresponding to (1'2'3'32 1)
     """
     dim = tpdm.shape[0]
@@ -760,19 +765,15 @@ def one_rdo_commutator_symm(two_body_tensor: np.ndarray,
 
     Args:
         two_body_tensor: 4-tensor for the coefficients of A
+
         tpdm: spin-orbital two-RDM p^ q^ r s corresponding to (1'2'2 1)
     """
     dim = tpdm.shape[0]
-    tensor_of_expectation = np.zeros(tuple([dim] * 2), dtype=tpdm.dtype)
     k2 = two_body_tensor.transpose(0, 1, 3, 2)
-    for p, q in product(range(dim), repeat=2):
-        commutator_expectation = 0.
-        #   (   2.00000) k2(p,a,b,c) cre(b) cre(c) des(q) des(a)
-        commutator_expectation += 2.0 * np.einsum('abc,bca', k2[p, :, :, :],
-                                                  tpdm[:, :, q, :])
+    tensor_of_expectation = np.zeros(tuple([dim] * 2), dtype=tpdm.dtype)
+    #   (   2.00000) k2(p,a,b,c) cre(b) cre(c) des(q) des(a)
+    tensor_of_expectation += 2.0 * np.einsum('pabc,bcqa->pq', k2, tpdm)
 
-        #   (  -2.00000) k2(q,a,b,c) cre(p) cre(a) des(b) des(c)
-        commutator_expectation += -2.0 * np.einsum('abc,abc', k2[q, :, :, :],
-                                                   tpdm[p, :, :, :])
-        tensor_of_expectation[p, q] = commutator_expectation
+    #   (  -2.00000) k2(q,a,b,c) cre(p) cre(a) des(b) des(c)
+    tensor_of_expectation += -2.0 * np.einsum('qabc,pabc->pq', k2, tpdm)
     return tensor_of_expectation
