@@ -15,15 +15,14 @@
 """
 
 from operator import itemgetter
-from typing import Any, Generator, KeysView, List, Set, Tuple, TYPE_CHECKING
+from typing import Any, Dict, KeysView, List, Set, Tuple, TYPE_CHECKING
 
 import numpy
 
 from fqe.bitstring import lexicographic_bitstring_generator
-from fqe.bitstring import check_conserved_bits, count_bits
+from fqe.bitstring import count_bits
 
 if TYPE_CHECKING:
-    #Avoid circular imports and only import for type-checking
     from fqe import wavefunction
 
 
@@ -32,32 +31,36 @@ def alpha_beta_electrons(nele: int, m_s: int) -> Tuple[int, int]:
     the number of alpha and beta electrons in the system.
 
     Args:
-        nele (int) - number of electrons
-        m_s (int) - spin angular momentum on the z-axis
+        nele (int): number of electrons
+
+        m_s (int): spin angular momentum on the z-axis
 
     Return:
-        number of alpha electrons (int), number of beta electrons (int)
+        Tuple[int, int]: number of alpha and beta electrons
     """
     if nele < 0:
         raise ValueError('Cannot have negative electrons')
     if nele < abs(m_s):
         raise ValueError('Spin quantum number exceeds physical limits')
+    if (nele + m_s) % 2 != 0:
+        raise ValueError('Parity of spin quantum number and number of '
+                         'electrons is incompatible')
     nalpha = int(nele + m_s) // 2
     nbeta = nele - nalpha
     return nalpha, nbeta
 
 
-def reverse_bubble_list(arr: List[Any]) -> int:
-    """Bubble Sort algorithm to arrange a list so that the lowest value is
-    stored in 0 and the highest value is stored in len(arr)-1.  It is included
+def reverse_bubble_list(arr: List[List[Any]]) -> int:
+    """Bubble Sort algorithm to arrange a list so that the highest value is
+    stored in 0 and the lowest value is stored in len(arr)-1. It is included
     here in order to access the swap count.
 
     Args:
-        arr (list) - object to be sorted
+        arr (list): List of objects to be sorted. The objects are sorted with \
+            respect to the value of object[0]. The data are sorted inplace.
 
     Returns:
-        arr (list) - sorted
-        swap_count (int) - number of permutations to achieve the sort
+        (int): number of permutations to achieve the sort
     """
     larr = len(arr)
     swap_count = 0
@@ -81,11 +84,10 @@ def bubblesort(arr: List[Any]) -> int:
     here in order to access the swap count.
 
     Args:
-        arr (list) - object to be sorted
+        arr (list): object to be sorted in place.
 
     Returns:
-        arr (list) - sorted
-        swap_count (int) - number of permutations to achieve the sort
+        (int): number of permutations to achieve the sort
     """
     larr = len(arr)
     swap_count = 0
@@ -107,10 +109,11 @@ def configuration_key_union(*argv: KeysView[Tuple[int, int]]
     all configuration keys in the list
 
     Args:
-        *args (list[(int, int)]) - any number of configuration key lists to be joined
+        *args (list[(int, int)]): any number of configuration key lists to be \
+            joined
 
     Returns:
-        list[(int, int)] - a list of unique configuration keys found among all
+        list[(int, int)]: a list of unique configuration keys found among all \
             the passed arguments
     """
     keyunion: Set[Tuple[int, int]] = set()
@@ -124,30 +127,27 @@ def configuration_key_intersection(*argv: List[Tuple[int, int]]
     """Return the intersection of the passed configuration key lists.
 
     Args:
-        *args (list[(int, int)]) - any number of configuration key lists to be joined
+        *args (list[(int, int)]): any number of configuration key lists to be \
+            intersected
 
     Returns:
-        list [(int, int)] - a list of configuration keys found in every
+        list[(int, int)]: a list of configuration keys found in every \
             configuration passed.
     """
-    keyinter = argv[0]
-    ref = []
+    ref = set(argv[0])
     for config in argv[1:]:
-        for key in config:
-            if key in keyinter:
-                ref.append(key)
-        keyinter = ref
-    return keyinter
+        ref = ref.intersection(config)
+    return list(ref)
 
 
 def init_bitstring_groundstate(occ_num: int) -> int:
     """Occupy the n lowest orbitals of a state in the bitstring representation
 
     Args:
-        occ_num (integer) - number of orbitals to occupy
+        occ_num (integer): number of orbitals to occupy
 
     Returns:
-        (integer) - bitstring representation of the ground state
+        (integer): bitstring representation of the ground state
     """
     return (1 << occ_num) - 1
 
@@ -156,61 +156,25 @@ def init_qubit_vacuum(nqubits: int) -> numpy.ndarray:
     """Build the ground state wavefunction for an nqubit system.
 
     Args:
-        nqubits (integer) - The number of qubits in the qpu
+        nqubits (integer): The number of qubits in the qpu
 
     Returns:
-        numpy.array(dtype=numpy.complex64)
+        numpy.array(dtype=numpy.complex128): ground state in the qubit \
+            representation
     """
     ground_state = numpy.zeros(2**nqubits, dtype=numpy.complex128)
     ground_state[0] = 1.0 + 0.0j
     return ground_state
 
 
-def ltlt_index_generator(dim: int
-                        ) -> Generator[Tuple[int, int, int, int], None, None]:
-    """Generate index sets into a lower triangle, lower triangle matrix
-
-    Args:
-        dim (int) - the dimension of the array
-
-    Returns:
-        (int, int, int, int) - unique pointers into the compressed matrix
-    """
-    lim = dim
-    for i in range(lim):
-        for j in range(i + 1):
-            for k in range(i + 1):
-                if k == i:
-                    _ull = j + 1
-                else:
-                    _ull = k + 1
-                for lst in range(_ull):
-                    yield i, j, k, lst
-
-
-def invert_bitstring_with_mask(string: int, masklen: int) -> int:
-    """Invert a bitstring with a mask.
-
-    Args:
-        string (bitstring) - the bitstring to invert
-        masklen (int) - the value to mask the inverted bitstring to
-
-    Returns:
-        (bitstring) - a bitstring inverted up to the masking length
-    """
-    mask = (1 << masklen) - 1
-    return ~string & mask
-
-
 def paritysort_int(arr: List[int]) -> Tuple[int, List[int]]:
     """Move all even numbers to the left and all odd numbers to the right
 
     Args:
-        arr list[int] - a list of integers to be sorted
+        arr (list[int]): a list of integers to be sorted in place.
 
     Returns:
-        arr [list] - mutated in place
-        swap_count (int) - number of exchanges needed to complete the sorting
+        (int): number of exchanges needed to complete the sorting
     """
     larr = len(arr)
     parr = [[i % 2, i] for i in arr]
@@ -236,11 +200,10 @@ def paritysort_list(arr):
     """Move all even numbers to the left and all odd numbers to the right
 
     Args:
-        arr list[int] - a list of integers to be sorted
+        arr (list[int]): a list of integers to be sorted in place
 
     Returns:
-        arr [list] - mutated in place
-        swap_count (int) - number of exchanges needed to complete the sorting
+        (int): number of exchanges needed to complete the sorting
     """
     larr = len(arr)
     parr = [[i[0] % 2, i] for i in arr]
@@ -268,18 +231,19 @@ def qubit_particle_number_sector(nqubits: int,
     which have a definite particle number.
 
     Args:
-        nqubits (int) - the number of qubits in the qpu
-        pnum (int) - the number of particles to build vectors into
+        nqubits (int): the number of qubits in the qpu
+
+        pnum (int): the number of particles to build vectors into
 
     Returns:
-        list[numpy.array(dtype=numpy.complex64)]
+        list[numpy.array(dtype=numpy.complex128)]
     """
-    occ = numpy.array([0, 1], dtype=numpy.int)
-    uno = numpy.array([1, 0], dtype=numpy.int)
-    seed = init_bitstring_groundstate(pnum)
-    pn_set = lexicographic_bitstring_generator(seed, nqubits)
+    occ = numpy.array([0, 1], dtype=numpy.int32)
+    uno = numpy.array([1, 0], dtype=numpy.int32)
+    pn_set = lexicographic_bitstring_generator(pnum, nqubits)
     vectors = []
-    for orbocc in pn_set:
+    for iorbocc in pn_set:
+        orbocc = int(iorbocc)
         if orbocc & 1:
             vec = occ
         else:
@@ -301,16 +265,17 @@ def qubit_config_sector(nqubits: int, pnum: int,
     which have a definite particle number and spin.
 
     Args:
-        nqubits (int) - the number of qubits in the qpu
-        pnum (int) - the number of particles to build vectors into
-        m_s (int) - the s_z spin quantum number
+        nqubits (int): the number of qubits in the qpu
+
+        pnum (int): the number of particles to build vectors into
+
+        m_s (int): the s_z spin quantum number
 
     Returns:
-        list[numpy.array(dtype=numpy.complex64)]
+        list[numpy.array(dtype=numpy.complex128)]
     """
-    occ = numpy.array([0, 1], dtype=numpy.int)
-    uno = numpy.array([1, 0], dtype=numpy.int)
-    seed = init_bitstring_groundstate(pnum)
+    occ = numpy.array([0, 1], dtype=numpy.int32)
+    uno = numpy.array([1, 0], dtype=numpy.int32)
     achk = 0
     bchk = 0
     pn_set = []
@@ -321,8 +286,9 @@ def qubit_config_sector(nqubits: int, pnum: int,
         else:
             achk += 2**num
 
-    initpn = lexicographic_bitstring_generator(seed, nqubits)
-    for occu in initpn:
+    initpn = lexicographic_bitstring_generator(pnum, nqubits)
+    for ioccu in initpn:
+        occu = int(ioccu)
         if (count_bits(occu & achk) - count_bits(occu & bchk)) == m_s:
             pn_set.append(occu)
 
@@ -348,17 +314,18 @@ def qubit_particle_number_index(nqubits: int, pnum: int) -> List[int]:
     with a specific particle number
 
     Args:
-        nqubits (int) - the number of qubits to act upon
-        pnum (int) - the number of particles to view
+        nqubits (int): the number of qubits to act upon
+
+        pnum (int): the number of particles to view
 
     Returns:
-        list[int] - integers indicating where in the qubit wavefunction the
+        list[int]: integers indicating where in the qubit wavefunction the
             basis state corresponds to particle number
     """
-    seed = init_bitstring_groundstate(pnum)
     indexes = []
-    pn_set = lexicographic_bitstring_generator(seed, nqubits)
-    for orbocc in pn_set:
+    pn_set = lexicographic_bitstring_generator(pnum, nqubits)
+    for occ in pn_set:
+        orbocc = int(occ)
         if orbocc & 1:
             index = 1
         else:
@@ -380,18 +347,19 @@ def qubit_particle_number_index_spin(nqubits: int,
     with a specific particle number and spin
 
     Args:
-        nqubits (int) - the number of qubits to act upon
-        pnum (int) - the number of particles to view
+        nqubits (int): the number of qubits to act upon
+
+        pnum (int): the number of particles to view
 
     Returns:
-        list[(int int)] - tuples of integers indicating where in the qubit
+        list[(int int)]: tuples of integers indicating where in the qubit
             wavefunction the basis state corresponds to particle number and
             return the corresponding spin
     """
-    seed = init_bitstring_groundstate(pnum)
     indexes = []
-    pn_set = lexicographic_bitstring_generator(seed, nqubits)
-    for orbocc in pn_set:
+    pn_set = lexicographic_bitstring_generator(pnum, nqubits)
+    for occ in pn_set:
+        orbocc = int(occ)
         totspn = 0
         curspn = -1
         if orbocc & 1:
@@ -416,23 +384,29 @@ def rand_wfn(adim: int, bdim: int) -> numpy.ndarray:
     """Utility for generating random normalized wavefunctions.
 
     Args:
-        dim (int) - length of the wavefunction
-        sparse (string) - a string indicating an approximate measure of how
-            often zeros should show up in the wavefunction
+        adim (int): length of the alpha string
+
+        bdim (int): length of the beta string
 
     Returns:
-        numpy.array(dim=dim, dtype=numpy.complex64)
+        numpy.ndarray(shape=(adim, bdim), dtype=numpy.complex128)
     """
     wfn = numpy.random.randn(adim, bdim).astype(numpy.complex128) + \
           numpy.random.randn(adim, bdim).astype(numpy.complex128)*1.j
 
-    norm = numpy.sqrt(numpy.vdot(wfn.flatten(), wfn.flatten()))
-    wfn /= norm
     return wfn
 
 
-def map_broken_symmetry(s_z, norb):
+def map_broken_symmetry(s_z: int,
+                        norb: int) -> Dict[Tuple[int, int], Tuple[int, int]]:
     """Create a map between spin broken and number broken wavefunctions.
+    Args:
+        s_z (int): quantum number s_z
+
+        norb (int): number of orbitals
+
+    Returns:
+        Dict[Tuple[int, int], Tuple[int, int]]
     """
     spin_to_number = {}
     nele = norb + s_z
@@ -455,10 +429,10 @@ def sort_configuration_keys(configs: KeysView[Tuple[int, int]]
     and then by the spin quantum number.
 
     Args:
-        wfn list[(int, int)] - a dictionary of keys
+        configs (KeysView[Tuple[int, int]]): a dictionary of keys
 
     Returns:
-        list with the sorted keys
+        List[Tuple[int, int]]: list with the sorted keys
     """
     return sorted(configs, key=itemgetter(0, 1))
 
@@ -467,12 +441,14 @@ def validate_config(nalpha: int, nbeta: int, norb: int) -> None:
     """ Check that the parameters passed are valid to build a configuration
 
     Args:
-        nalpha (int) - the number of alpha electrons
-        nbeta (int) - the number of beta electrons
-        norb (int) - the number of spatial orbitals
+        nalpha (int): the number of alpha electrons
+
+        nbeta (int): the number of beta electrons
+
+        norb (int): the number of spatial orbitals
 
     Returns:
-        nothing - only raises errors if necessary
+        nothing: only raises errors if necessary
     """
     if nalpha < 0:
         raise ValueError("Cannot have negative number of alpha electrons")
@@ -487,9 +463,13 @@ def validate_config(nalpha: int, nbeta: int, norb: int) -> None:
         raise ValueError("Insufficient number of orbitals")
 
 
-def validate_tuple(matrices) -> None:
+def validate_tuple(matrices: Tuple[numpy.ndarray, ...]) -> None:
     """Validate that the tuple passed in is valid for initializing a general
     Hamiltonian
+
+    Args:
+        matrices(tuple[Nparray,...]): arrays to be used to initialize a \
+            Hamiltonian
     """
     assert isinstance(matrices, tuple)
     for rank, term in enumerate(matrices):
@@ -503,13 +483,14 @@ def dot(wfn1: 'wavefunction.Wavefunction',
     not use the conjugate.  See vdot for the similar conjugate functionality.
 
     Args:
-        wfn1 (wavefunction.Wavefunction) - wavefunction corresponding to the
+        wfn1 (wavefunction.Wavefunction): wavefunction corresponding to the
             row vector
-        wfn2 (wavefunction.Wavefunction) - wavefunction corresponding to the
+
+        wfn2 (wavefunction.Wavefunction): wavefunction corresponding to the
             coumn vector
 
     Returns:
-        (complex) - scalar as result of the dot product
+        (complex): scalar as result of the dot product
     """
     brakeys = wfn1.sectors()
     ketkeys = wfn2.sectors()
@@ -517,8 +498,8 @@ def dot(wfn1: 'wavefunction.Wavefunction',
     ipval = .0 + .0j
     for sector in keylist:
         ipval += numpy.dot(
-            wfn1.get_coeff(sector).flatten(),
-            wfn2.get_coeff(sector).flatten())
+            wfn1.get_coeff(sector).flat,
+            wfn2.get_coeff(sector).flat)
     return ipval
 
 
@@ -528,13 +509,14 @@ def vdot(wfn1: 'wavefunction.Wavefunction',
     the elements of wfn1.
 
     Args:
-        wfn1 (wavefunction.Wavefunction) - wavefunction corresponding to the
+        wfn1 (wavefunction.Wavefunction): wavefunction corresponding to the
             conjugate row vector
-        wfn2 (wavefunction.Wavefunction) - wavefunction corresponding to the
+
+        wfn2 (wavefunction.Wavefunction): wavefunction corresponding to the
             coumn vector
 
     Returns:
-        (complex) - scalar as result of the dot product
+        (complex): scalar as result of the dot product
     """
     brakeys = wfn1.sectors()
     ketkeys = wfn2.sectors()
@@ -542,29 +524,32 @@ def vdot(wfn1: 'wavefunction.Wavefunction',
     ipval = .0 + .0j
     for config in keylist:
         ipval += numpy.vdot(
-            wfn1.get_coeff(config).flatten(),
-            wfn2.get_coeff(config).flatten())
+            wfn1.get_coeff(config).flat,
+            wfn2.get_coeff(config).flat)
     return ipval
 
 
-def zero_transform(string0: int, unocc: int, occ: int, norb: int) -> bool:
-    """Given a bitstring, determine if it satisfies the occupation and
-    nonoccupation conditions necessary to be non zero when a product of creation
-    and annihilation operators are applied.
+def tensors_equal(tensor1: Dict[int, numpy.ndarray], \
+                    tensor2: Dict[int, numpy.ndarray]) -> bool:
+    """ Comparison function for two tensors (dictionaries as used in
+    subclasses of Hamiltonian). Used in implementing equality operators
+    in Hamiltonian subclasses.
+
+    Note: it compares using numpy.all and not numpy.isclose, so that even small
+    numerical deviations may lead to inequality.
 
     Args:
-        string0 (bitstring) - the occupation representation being acted upon
-        unocc (bitstring) - orbitals which should be unoccupied in string0
-        occ (bitstring) - orbitals which should be occupied in string0
-        norb (int) - the number of spatial orbitals for masking the bitstrings
+        tensor1 (Dict[int, numpy.ndarray]): first tensor
+
+        tensor2 (Dict[int, numpy.ndarray]): second tensor
 
     Returns:
-        (bool) - true if the transformation is non zero, false if the
-            transformation is zero
+        bool: True if all keys and corresponding values of tensor1 match to that
+        of tensor2, false otherwise.
     """
-    if check_conserved_bits(string0, occ):
-        if check_conserved_bits(invert_bitstring_with_mask(string0, norb),
-                                unocc):
+    if tensor1.keys() != tensor2.keys():
+        return False
+    for (k, v), (k2, v2) in zip(tensor1.items(), tensor2.items()):
+        if k != k2 or not numpy.all(v == v2):
             return False
-
     return True
